"""
.. module:: world
    :platform: Linux, Unix, Windows
    :synopsis: Provides classes for representing a world and its edges.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""
        
import json
        

class World:
    def __init__(self, edges=None):
        """Create a new world consisting of the given *edges*.
        
        :param list edges: a list of :class:`Edge`s

        """
        self._nodes = set()
        self.edges = {}
        if edges is not None:
            for e in edges:
                self.add_edge(e)
                
    @property
    def nodes(self):
        return list(self._nodes)
    
    def reset_pheromone(self, level=None):
        """Reset the amount of pheromone on every edge to *level*.
        """
        level = level or 0.01
        for edge in self.edges.values():
            edge.pheromone = level
        
    def add_edge(self, edge):
        """Add *edge* to the :class:`World`.
        
        :param :class:`Edge` edge: the :class:`Edge` to add
        
        """
        if not isinstance(edge, Edge):
            raise TypeError("edge must be <type Edge>")
        self._nodes.add(edge.start)
        self._nodes.add(edge.end)
        self.edges[edge.start, edge.end] = edge    
        
    

class Edge:
    """This class represents the link connecting two nodes.

    In addition to start and end nodes, every :class:`Edge` has a legnth and a
    pheromone level.  The length represents static, apriori information, while
    the pheromone level represents dynamic, posteriori information.

    """
    def __init__(self, a, b, length=None, pheromone=None):
        """Create a new :class:`Edge` between *a* and *b*.

        :param Node a: the node at the start of the :class:`Edge`
        :param Node b: the node at the end of the :class:`Edge`
        :param float length: the length of the :class:`Edge` (default=1)
        :param float pheromone: the amount of pheromone on the :class:`Edge` 
                                (default=0.1)

        """
        self.start = a
        self.end = b
        self.length = 1 if length is None else length
        self.pheromone = 0.1 if pheromone is None else pheromone
        
    def __repr__(self):
        return "Edge from {} to {} that is {} long.".format(
                self.start, self.end, self.length)
        
    def __eq__(self, other):
        """Return ``True`` iff *other* has identical properties.

        Two :class:`Edge`s are **not** equal if their lengths or pheromone
        level differs.

        :rtype: bool

        """
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False
        
        
class Node:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._hash = None
        
    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(self.__dict__.items()))
        return self._hash
            
    def __bool__(self):
        return True
      
    # 
    def __repr__(self):
        return json.dumps(
            {k:v for k,v in self.__dict__.items() if not k.startswith("_")},
            default=str)
        
    # Returns true iff other is the same type and has identical "public"
    # properties.
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            sd = {k:v for k, v in self.__dict__.items() if not k.startswith("_")}
            od = {k:v for k, v in other.__dict__.items() if not k.startswith("_")}
            return sd == od
        return False


"""
.. module:: ant
    :platform: Linux, Unix, Windows
    :synopsis: Provides functionality for finding each solution step as well
               as representing a complete solution.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

from .world import World
import itertools
import random
import bisect

class Ant:
    """
    A single independent finder of solutions to the world.

    """
    uid = 0

    def __init__(self, world, alpha=1, beta=3, start=None):
        """
        Create a new Ant for the given world.

        :param World world: the world in which to travel
        :param float alpha: the relative importance of pheromone (default=1)
        :param float beta: the relative importance of distance (default=3)
        :param Node start: the starting node (default=None)

        """
        Ant.uid = self.uid = Ant.uid + 1
        self.world = world
        self.alpha = alpha
        self.beta = beta
        self.trip_complete = False
        self.distance = 0
        self.reset(start)

    @property
    def moves(self):
        """
        Return a list of the moves made so far
        
        :returns: A list of (start, end) tuples
        :rtype: list
        
        """
        # TODO: Decide whether to simply return edges.
        p, n = self.path, len(self.path)
        return [(p[i], p[(i + 1) % n]) for i in range(n)]
        
    def clone(self):
        """
        Return a new :class:`Ant` with exactly the same property values.

        Note that unlike copy, this method preserves even the UID of an Ant.

        :returns: a clone
        :rtype: :class:`Ant`

        """
        ant = Ant(self.world, self.alpha, self.beta, self.start)
        ant.uid = self.uid
        ant.node = self.node
        ant.path = self.path[:]
        ant.distance = self.distance
        return ant

    def __lt__(self, other):
        return self.distance < other.distance

    def reset(self, start=None):
        """
        Reset everything so that a new solution can be found.

        Note that calling this method destroys the previous path, moves, and
        distance traveled by the ant.
        
        :param Node start: which :class:`Node` to start the next solution from
                           (by default, the previous starting :class:`Node` is
                            used again)
        
        """
        self.start = start
        self.node = self.start
        self.distance = 0
        self.path = []
        if start is not None:
            self.path.append(start)

    def can_move(self):
        """
        Return true if there are more than zero possible moves to make.

        """
        return len(self.path) < len(self.world.nodes)

    def move(self):
        """
        Choose a valid move and make it.

        """
        moves = self.get_possible_moves()
        move = self.choose_move(moves)
        if move:
            move_made = (self.node, move)
            self.make_move(move)
            return move_made
        return None

    def get_possible_moves(self):
        """
        Return the set of all moves that can currently be made.

        :returns: all currently possible moves
        :rtype: set
        
        """
        return set(self.world.nodes) - set(self.path)

    def choose_move(self, moves):
        """
        Return which move to make by choosing from *moves*.
        
        :param list moves: a list of all possible moves
        
        :returns: the move to make
        :rtype: :class:`Node`
        
        """
        N = len(moves)
        if N == 0:
            return None     # No more moves
        
        # Find the individual weight of each move.
        moves = list(moves) # it may be given as a set, but we need order
        weights = []
        if self.node is None:
            weights = [1 for i in range(N)]
        else:
            for m in moves:
                e = self.world.edges[self.node, m]
                weights.append(self.calculate_weight(e))
        
        # Ensure the index of the last element is never exceeded.
        total_weight = sum(weights) or 1
        cumdist = list(itertools.accumulate(weights)) + [total_weight]
        
        # Choose a random place within the distrution of weights and return the
        # move at the index of would-be insertion.
        r = random.random() * total_weight
        i = bisect.bisect(cumdist, r)
        return moves[i]
        
    def calculate_weight(self, e):
        """
        Calculate the weight of the given :class:`Edge` *e*.
        
        :param Edge e: the edge of which the weight shall be calculated

        :returns: the weight (projected usefullness) of *e*
        :rtype: float
        
        """
        return e.pheromone ** self.alpha * (1 / (e.length or 1)) ** self.beta

    def make_move(self, move):
        """
        Perform the given *move*.
        
        :param Node move: the :class:`Node` to move to
        
        """
        self.path.append(move)
        n = len(self.path)
        
        # If moving to the starting node, no distance was actually traveled.
        if n == 1:
            self.start = move        
        else:
            self.distance += self.world.edges[self.node, move].length
        
        # If moving to the last node, add the distance from the last node back
        # to the first node to "complete" the path.
        if n == len(self.world.nodes):
            final = self.world.edges[self.path[-1], self.path[0]].length
            self.distance += final

        self.node = move



from ..world import World, Edge, Node
import unittest
import math

class WorldTest(unittest.TestCase):
    def setUp(self):
        self.nodeA = Node(name="A")
        self.nodeB = Node(name="B")
        self.nodeC = Node(name="C")
        self.edgeAB = Edge(self.nodeA, self.nodeB)
        self.edgeBA = Edge(self.nodeB, self.nodeA)
        self.edgeAC = Edge(self.nodeA, self.nodeC)
        self.edgeCA = Edge(self.nodeC, self.nodeA)
        self.edgeBC = Edge(self.nodeB, self.nodeC)
        self.edgeCB = Edge(self.nodeC, self.nodeB)
        self.edgeAB2 = Edge(self.nodeA, self.nodeB,
            length=self.edgeAB.length + 1)

class WorldCreation(WorldTest):
    def test_world_creation_with_no_arguments(self):
        w = World()
        self.assertEqual(w.nodes, [])
        self.assertEqual(w.edges, {})
        
    def test_world_creation_from_edge_list(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        self.assertEqual(len(w.nodes), 2)
        self.assertTrue(self.nodeA in w.nodes)
        self.assertTrue(self.nodeB in w.nodes)

    def test_world_creation_failure_from_list_of_nonedges(self):
        not_edges = {"Not", "an", "edge", "list"}
        self.assertRaises(TypeError, World, (not_edges,))
    
    def test_world_creation_failure_from_extra_kwargs(self):
        self.assertRaises(TypeError, World, unknown="keyword")
        
    def test_world_creation_failure_from_extra_args(self):
        self.assertRaises(TypeError, World, (1,2,3))
    
    
class EdgeAddition(WorldTest):
    def test_world_correct_edge_count(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeBA)
        self.assertEqual(len(w.edges), 2, "Wrong number of edges added")

    def test_world_no_duplicate_nodes(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeAC)
        self.assertEqual(len(w.nodes), 3, "Wrong number of nodes extracted")
        
    def test_world_edge_replacement(self):
        w = World()
        w.add_edge(self.edgeAB)
        w.add_edge(self.edgeAB2)
        self.assertEqual(len(w.edges), 1)
        self.assertEqual(w.edges[self.nodeA, self.nodeB], self.edgeAB2)
        

class PheromoneReset(WorldTest):
    def test_world_pheromone_level_of_every_edge_reset(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        p = 1234
        w.reset_pheromone(p)
        self.assertTrue(all(e.pheromone == p for e in w.edges.values()))
        
    def test_world_pheromone_can_never_be_zero(self):
        edges = [self.edgeAB, self.edgeBA]
        w = World(edges)
        p = 0
        w.reset_pheromone(p)
        self.assertTrue(all(e.pheromone > 0 for e in w.edges.values()))
        

class EdgeTest(unittest.TestCase):
    def setUp(self):
        self.nodeA = Node(name="A")
        self.nodeB = Node(name="B")
        self.nodeC = Node(name="C")


class EdgeCreation(EdgeTest):        
    def test_edge_creation_with_no_args(self):
        self.assertRaises(TypeError, Edge)
        
    def test_edge_creation_with_mutable_nodes(self):
        self.assertRaises(TypeError, Edge, ({}, {}))
        
    def test_edge_creation_with_immutable_nodes(self):
        e = Edge(self.nodeA, self.nodeB)
        self.assertEqual(e.start, self.nodeA)
        self.assertEqual(e.end, self.nodeB)
        
    def test_edge_creation_with_custom_length(self):
        length = 100
        e = Edge(self.nodeA, self.nodeB, length=length)
        self.assertEqual(e.length, length)
        
    def test_edge_creation_with_custom_pheromone_level(self):
        level = 100
        e = Edge(self.nodeA, self.nodeB, pheromone=level)
        self.assertEqual(e.pheromone, level)


class EdgeEquality(EdgeTest):
    def test_edge_identical_edge_equality(self):
        e1 = Edge(self.nodeA, self.nodeB)
        e2 = Edge(self.nodeA, self.nodeB)
        self.assertEqual(e1, e2)
        
    def test_edge_length_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB, length=1)
        e2 = Edge(self.nodeA, self.nodeB, length=2)
        self.assertNotEqual(e1, e2)
        
    def test_edge_pheromone_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB, pheromone=1)
        e2 = Edge(self.nodeA, self.nodeB, pheromone=2)
        self.assertNotEqual(e1, e2)
        
    def test_edge_node_difference_inequality(self):
        e1 = Edge(self.nodeA, self.nodeB)
        e2 = Edge(self.nodeB, self.nodeA)
        self.assertNotEqual(e1, e2)
        

class NodeTest(unittest.TestCase):
    def setUp(self):
        self.dataA = {'name': "A"}
        self.dataB = {'name': "B", 'height': None}
        self.dataC = {'name': "C", 'age': 27}
        

class NodeCreation(NodeTest):
    def test_node_creation_with_no_args(self):
        n = Node()
        self.assertEqual(list(n.__dict__.keys()), ['_hash'])
        
    def test_node_creation_with_keyword_args(self):
        for d in (self.dataA, self.dataB, self.dataC):
            n = Node(**d)
            self.assertTrue(all(n.__dict__[k] == d[k] for k in d))
            
    def test_node_creation_with_positional_args(self):
        self.assertRaises(TypeError, Node, (1,))
        

class NodeHashes(NodeTest):
    def test_node_hash_consistency(self):
        h = hash(Node())
        self.assertTrue(all(hash(Node()) == h for _ in range(10)))
        
    def test_node_hash_uniqueness(self):
        nodes = [Node(**d) for d in (self.dataA, self.dataB, self.dataC)]
        hashes = list(map(hash, nodes))
        for i in range(len(hashes)):
            for j in range(len(hashes)):
                if i != j:
                    self.assertNotEqual(hashes[i], hashes[j])
    
    def test_node_use_as_dict_key(self):
        nodeA = Node(**self.dataA)
        nodeB = Node(**self.dataB)
        node_dict = {}
        node_dict[nodeA] = 'a'
        node_dict[nodeB] = 'b'
        self.assertTrue(nodeA in node_dict)
        self.assertTrue(nodeB in node_dict)
        self.assertEqual(node_dict[nodeA], 'a')
        self.assertEqual(node_dict[nodeB], 'b')
        

class NodeEquality(NodeTest):
    def test_node_equality(self):
        a = Node(**self.dataA)
        b = Node(**self.dataA)
        self.assertEqual(a, b)
        
    def test_node_inequality(self):
        a = Node(**self.dataA)
        b = Node(**self.dataB)
        self.assertNotEqual(a, b)
        
    def test_node_inequality_with_same_attrs(self):
        a = Node(x=3)
        b = Node(x=4)
        self.assertNotEqual(a, b)
        
    def test_node_inequality_with_same_values(self):
        a = Node(x=3)
        b = Node(y=3)
        self.assertNotEqual(a, b)
        
    def test_node_indexability(self):
        node_list = [Node(**d) for d in (self.dataA, self.dataB, self.dataC)]
        for i in range(len(node_list)):
            self.assertEqual(node_list.index(node_list[i]), i)
            
            
if __name__ == '__main__':
    unittest.main()


"""A Python3 implementation of the Ant Colony Optimization Meta-Heuristic.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

from .ant import Ant 
from .world import World, Edge, Node
from .solver import Solver
